@@ -4,11 +4,13 @@ use 5.010; # Uses // defined-or operator
 use strict;
 use warnings FATAL => 'all';
 
+use Capture::Tiny 'capture';
 use Carp;
 use DBI;
 use File::Temp;
-use IO::File;
-use Capture::Tiny 'capture';
+use Log::Any;
+use Scope::Guard qw(guard);
+
 use Moo;
 use namespace::clean;
 
@@ -19,11 +21,11 @@ PGObject
 
 =head1 VERSION
 
-Version 0.130.1
+version 1.4.0
 
 =cut
 
-our $VERSION = '0.130.1';
+our $VERSION = '1.4.0';
 
 
 =head1 SYNOPSIS
@@ -32,11 +34,13 @@ This module provides an interface to the basic Postgres db manipulation
 utilities.
 
  my $db = PGObject::Util::DBAdmin->new(
-    username => 'postgres',
-    password => 'mypassword',
-    host     => 'localhost',
-    port     => '5432',
-    dbname   => 'mydb'
+    connect_data => {
+       user     => 'postgres',
+       password => 'mypassword',
+       host     => 'localhost',
+       port     => '5432',
+       dbname   => 'mydb'
+    }
  );
 
  my @dbnames = $db->list_dbs(); # like psql -l
@@ -48,9 +52,102 @@ utilities.
 
  my $db2 = PGObject::Util::DBAdmin->new($db->export, (dbname => 'otherdb'));
 
+ my $db3 = PGObject::Util::DBAdmin->new(
+    connect_data => {
+       service     => 'zephyr',
+       sslmode     => 'require',
+       sslkey      => "$HOME/.postgresql/postgresql.key",
+       sslcert     => "$HOME/.postgresql/postgresql.crt",
+       sslpassword => 'your-sslpassword',
+    }
+ );
+
+
 =head1 PROPERTIES
 
-=head2 username
+=head2 connect_data
+
+Contains a hash with connection parameters; see L<the PostgreSQL
+documentation|https://www.postgresql.org/docs/current/libpq-connect.html#LIBPQ-PARAMKEYWORDS>
+for supported parameters.
+
+The usual parameters are:
+
+=over
+
+=item * user
+
+=item * password
+
+=item * dbname
+
+=item * host
+
+=item * port
+
+=back
+
+Please note that the key C<requiressl> is deprecated in favor of
+C<sslmode> and isn't supported.
+
+=cut
+
+# Not supported
+#  PGSERVICEFILE: (because no connect string equiv)
+#  PGREQUIRESSL: deprecated
+my %connkey_env = qw(
+    host                     PGHOST
+    hostaddr                 PGHOSTADDR
+    dbname                   PGDATABASE
+    user                     PGUSER
+    password                 PGPASSWORD
+    passfile                 PGPASSFILE
+    channel_binding          PGCHANNELBINDING
+    service                  PGSERVICE
+    options                  PGOPTIONS
+    sslmode                  PGSSLMODE
+    sslcompression           PGSSLCOMPRESSION
+    sslcert                  PGSSLCERT
+    sslkey                   PGSSLKEY
+    sslrootcert              PGSSLROOTCERT
+    sslcrl                   PGSSLCRL
+    requirepeer              PGREQUIREPEER
+    ssl_min_protocol_version PGSSLMINPROTOCOLVERSION
+    ssl_max_protocol_version PGSSLMAXPROTOCOLVERSION
+    gssencmode               PGGSSENCMODE
+    krbsrvname               PGKRBSRVNAME
+    gsslib                   PGGSSLIB
+    connect_timeout          PGCONNECT_TIMEOUT
+    client_encoding          PGCLIENTENCODING
+    target_session_attrs     PGTARGETSESSIONATTRS
+    );
+my @connstr_keys = ((grep { not ($_ eq 'user' or $_ eq 'password') }
+                     keys %connkey_env),
+                    qw(application_name fallback_application_name
+                    keepalives keepalives_idle keepalives_interval
+                    keepalives_count tcp_user_timeout replication sslpassword),
+    );
+
+sub _connect_data_env {
+    my ($connect_data) = @_;
+    my @keys = grep { exists $connkey_env{$_}
+                      and defined $connect_data->{$_} } keys %$connect_data;
+    return map { $connkey_env{$_} => $connect_data->{$_} } @keys;
+}
+
+sub _connect_data_str {
+    my ($connect_data) = @_;
+    my @keys = grep { defined $connect_data->{$_} } @connstr_keys;
+    return join(';', map {
+        my $val = $connect_data->{$_};
+        $val =~ s/\\/\\\\/g;
+        $val =~ s/'/\\'/g;
+        "$_='$val'"; } @keys );
+}
+
+has connect_data => (is => 'ro');
+
+=head2 username (deprecated)
 
 The username used to authenticate with the PostgreSQL server.
 
@@ -58,7 +155,7 @@ The username used to authenticate with the PostgreSQL server.
 
 has username => (is => 'ro');
 
-=head2 password
+=head2 password (deprecated)
 
 The password used to authenticate with the PostgreSQL server.
 
@@ -66,7 +163,7 @@ The password used to authenticate with the PostgreSQL server.
 
 has password => (is => 'ro');
 
-=head2 host
+=head2 host (deprecated)
 
 In PostgreSQL, this can refer to the hostname or the absolute path to the
 directory where the UNIX sockets are set up.
@@ -75,7 +172,7 @@ directory where the UNIX sockets are set up.
 
 has host => (is => 'ro');
 
-=head2 port
+=head2 port (deprecated)
 
 Default '5432'
 
@@ -83,7 +180,7 @@ Default '5432'
 
 has port => (is => 'ro');
 
-=head2 dbname
+=head2 dbname (deprecated)
 
 The database name to create or connect to.
 
@@ -111,35 +208,130 @@ notes in L</"CAPTURING">.
 
 has stdout => (is => 'ro');
 
+=head2 logger
+
+Provides a reference to the logger associated with the current instance. The
+logger uses C<ref $self> as its category, eliminating the need to create
+new loggers when deriving from this class.
+
+If you want to override the logger-instantiation behaviour, please implement
+the C<_build_logger> builder method in your derived class.
+
+=cut
+
+has logger => (is => 'ro', lazy => 1, builder => '_build_logger');
+
+sub _build_logger {
+    return Log::Any->get_logger(category => ref $_[0]);
+}
+
+
+our %helpers =
+    (
+     create => [ qw/createdb/ ],
+     run_file => [ qw/psql/ ],
+     backup => [ qw/pg_dump/ ],
+     backup_globals => [ qw/pg_dumpall/ ],
+     restore => [ qw/pg_restore psql/ ],
+     drop => [ qw/dropdb/ ],
+    );
+
+=head1 GLOBAL VARIABLES
+
+
+=head2 %helper_paths
+
+This hash variable contains as its keys the names of the PostgreSQL helper
+executables C<psql>, C<dropdb>, C<pg_dump>, etc. The values contain the
+paths at which the executables to be run are located. The default values
+are the names of the executables only, allowing them to be looked up in
+C<$PATH>.
+
+Modification of the values in this variable are the strict realm of
+I<applications>. Libraries using this library should defer potential
+required modifications to the applications based upon them.
+
+=cut
+
+our %helper_paths =
+    (
+     psql => 'psql',
+     dropdb => 'dropdb',
+     createdb => 'createdb',
+     pg_dump => 'pg_dump',
+     pg_dumpall => 'pg_dumpall',
+     pg_restore => 'pg_restore',
+    );
+
+sub _run_with_env {
+    my %args = @_;
+    my $env = $args{env};
+
+    local %ENV = (
+        # Note that we're intentionally *not* passing
+        # PERL5LIB & PERL5OPT into the environment here!
+        # doing so prevents the system settings to be used, which
+        # we *do* want. If we don't, hopefully, that's coded into
+        # the executables themselves.
+        # Before using this whitelisting, coverage tests in LedgerSMB
+        # would break on the bleeding through this caused.
+        HOME => $ENV{HOME},
+        PATH => $ENV{PATH},
+        %{$env // {}},
+        );
+
+    return system @{$args{command}};
+}
 
 sub _run_command {
     my ($self, %args) = @_;
     my $exit_code;
+    my %env = (
+        # lowest priority: existing environment variables
+        (map { $ENV{$_} ? ($_ => $ENV{$_})  : () }
+         qw(PGUSER PGPASSWORD PGHOST PGPORT PGDATABASE PGSERVICE)),
+        # overruled by middle priority: object connection parameters
+        _connect_data_env($self->connect_data),
+        # overruled by highest priority: specified environment
+        ($args{env}   ? %{$args{env}} : ()),
+        );
+    $self->logger->debugf(
+        sub {
+            return 'Running with environment: '
+                . join(' ', map { qq|$_="$env{$_}"| } keys %env );
+        });
 
     # Any files created should be accessible only by the current user
     my $original_umask = umask 0077;
+    {
+        my $guard = guard { umask $original_umask; };
 
-    ($self->{stdout}, $self->{stderr}, $exit_code) = capture {
-        system @{$args{command}};
-    };
-
-    if(defined ($args{errlog} // $args{stdout_log})) {
-        $self->_write_log_files(%args);
+        ($self->{stdout}, $self->{stderr}, $exit_code) = capture {
+            _run_with_env(%args, env => \%env);
+        };
+        if(defined ($args{errlog} // $args{stdout_log})) {
+            $self->_write_log_files(%args);
+        }
     }
 
-    # Reset original umask
-    umask $original_umask;
-
-    # Return true if system command is successful
-    return ($exit_code == 0);
-}
-
-
-sub _unlink_file_and_croak {
-    my ($self, $filename, $message) = @_;
-
-    unlink $filename or carp "error unlinking $filename $!";
-    croak $message;
+    if ($exit_code != 0) {
+        for my $filename (@{$args{unlink}}) {
+            unlink $filename or carp "error unlinking '$filename': $!";
+        }
+        my $command = join( ' ', map { "'$_'" } @{$args{command}} );
+        my $err;
+        if ($? == -1) {
+            $err = "$!";
+        }
+        elsif ($? & 127) {
+            $err = sprintf('died with signal %d', ($? & 127));
+        }
+        else {
+            $err = sprintf('exited with code %d', ($? >> 8));
+        }
+        croak "$args{error}; (command: $command): $err";
+    }
+    return 1;
 }
 
 
@@ -206,16 +398,112 @@ sub _append_to_file {
 
 Creates a new db admin object for manipulating databases.
 
+=head2 BUILDARGS
+
+Compensates for the legacy invocation with the C<username>, C<password>,
+C<host>, C<port> and C<dbname> parameters.
+
+=head2 verify_helpers( [ helpers => [...]], [operations => [...]])
+
+Verifies ability to execute (external) helper applications by
+method name (through the C<operations> argument) or by external helper
+name (through the C<helpers> argument). Returns a hash ref with each
+key being the name of a helper application (see C<helpers> below) with
+the values being a boolean indicating whether or not the helper can be
+successfully executed.
+
+Valid values in the array referenced by the C<operations> parameter are
+C<create>, C<run_file>, C<backup>, C<backup_globals>, C<restore> and
+C<drop>; the methods this module implements with the help of external
+helper programs. (Other values may be passed, but unsupported values
+aren't included in the return value.)
+
+Valid values in the array referenced by the C<helpers> parameter are the
+names of the PostgreSQL helper programs C<createdb>, C<dropdb>, C<pg_dump>,
+C<pg_dumpall>, C<pg_restore> and C<psql>. (Other values may be passed, but
+unsupported values will not be included in the return value.)
+
+When no arguments are passed, all helpers will be tested.
+
+Note: C<verify_helpers> is a class method, meaning it wants to be called
+as C<PGObject::Util::DBAdmin->verify_helpers()>.
+
+=cut
+
+around 'BUILDARGS' => sub {
+    my ($orig, $class, @args) = @_;
+
+    ## 1.1.0 compatibility code (allow a reference to be passed in)
+    my %args = (@args == 1 and ref $args[0]) ? (%{$args[0]}) : (@args);
+
+    # deprecated field support code block
+    if (exists $args{connect_data}) {
+        # Work-around for 'export' creating the expectation that
+        # parameters may be overridable; I've observed the pattern
+        #  ...->new($db->export, (dbname => 'newdb'))
+        # which we "solve" by hacking the dbname arg into the connect_data
+        # Don't overwrite connect_data, because it may be used elsewhere...
+        $args{connect_data} = {
+            %{$args{connect_data}},
+            dbname => ($args{dbname} // $args{connect_data}->{dbname})
+        };
+
+        # Now for legacy purposes hack the connection parameters into
+        # connect_data
+        $args{username} = $args{connect_data}->{user};
+        $args{$_}       = $args{connect_data}->{$_} for (qw(password dbname
+                                                         host port));
+    }
+    else {
+        $args{connect_data}             = {};
+        $args{connect_data}->{user}     = $args{username};
+        $args{connect_data}->{password} = $args{password};
+        $args{connect_data}->{dbname}   = $args{dbname};
+        $args{connect_data}->{host}     = $args{host};
+        $args{connect_data}->{port}     = $args{port};
+    }
+    return $class->$orig(%args);
+};
+
+
+
+
+sub _run_capturing_output {
+    my @args = @_;
+    my ($stdout, $stderr, $exitcode) = capture { _run_with_env(@args); };
+
+    return $exitcode;
+}
+
+sub verify_helpers {
+    my ($class, %args) = @_;
+
+    my @helpers = (
+        @{$args{helpers} // []},
+        map { @{$helpers{$_} // []} } @{$args{operations} // []}
+        );
+    if (not @helpers) {
+        @helpers = keys %helper_paths;
+    }
+    return {
+        map {
+            $_ => not _run_capturing_output(command =>
+                                            [ $helper_paths{$_} , '--help' ])
+        } @helpers
+    };
+}
+
+
 =head2 export
 
-Exports the database parameters in a hash so it can be used to create another
+Exports the database parameters as a list so it can be used to create another
 object.
 
 =cut
 
 sub export {
     my $self = shift;
-    return map {$_ => $self->$_() } qw(username password host port dbname)
+    return ( connect_data => $self->connect_data );
 }
 
 =head2 connect($options)
@@ -229,34 +517,33 @@ Connection options may be specified in the $options hashref.
 sub connect {
     my ($self, $options) = @_;
 
-    my $connect = 'dbname="' . $self->dbname . '"';
-
-    $connect .= ';host=' . $self->host
-        if defined $self->host;
-
-    $connect .= ';port=' . $self->port
-        if defined $self->port;
-
-    my $dbh = DBI->connect(
+    my $connect = _connect_data_str($self->connect_data);
+    my $dbh     = DBI->connect(
         'dbi:Pg:' . $connect,
-        $self->username,
-        $self->password,
+        $self->connect_data->{user} // '',    # suppress use of DBI_USER
+        $self->connect_data->{password} // '',# suppress use of DBI_PASS
         $options
-    ) or croak 'Could not connect to database!';
+    ) or croak 'Could not connect to database: ' . $DBI::errstr;
 
     return $dbh;
 }
 
-=head2 server_version
+=head2 server_version([$dbname])
 
 Returns a version string (like 9.1.4) for PostgreSQL. Croaks on error.
 
+When a database name is specified, uses that database to connect to,
+using the credentials specified in the instance.
+
+If no database name is specified, 'template1' is used.
+
 =cut
 
 sub server_version {
     my $self = shift @_;
+    my $dbname = (shift @_) || 'template1';
     my $version =
-           __PACKAGE__->new($self->export, (dbname => 'template1')
+           __PACKAGE__->new($self->export, (dbname => $dbname)
                            )->connect->selectrow_array('SELECT version()');
     my ($retval) = $version =~ /(\d+\.\d+\.\d+)/
         or croak 'failed to extract version string';
@@ -264,17 +551,23 @@ sub server_version {
 }
 
 
-=head2 list_dbs
+=head2 list_dbs([$dbname])
 
 Returns a list of db names.
 
+When a database name is specified, uses that database to connect to,
+using the credentials specified in the instance.
+
+If no database name is specified, 'template1' is used.
+
 =cut
 
 sub list_dbs {
     my $self = shift;
+    my $dbname = (shift @_) || 'template1';
 
     return map { $_->[0] }
-           @{ __PACKAGE__->new($self->export, (dbname => 'template1')
+           @{ __PACKAGE__->new($self->export, (dbname => $dbname)
            )->connect->selectall_arrayref(
                  'SELECT datname from pg_database order by datname'
            ) };
@@ -304,17 +597,13 @@ sub create {
     my $self = shift;
     my %args = @_;
 
-    local $ENV{PGPASSWORD} = $self->password if $self->password;
-
-    my @command = ('createdb');
-    defined $self->username and push(@command, '-U', $self->username);
+    my @command = ($helper_paths{createdb});
     defined $args{copy_of}  and push(@command, '-T', $args{copy_of});
-    defined $self->host     and push(@command, '-h', $self->host);
-    defined $self->port     and push(@command, '-p', $self->port);
-    defined $self->dbname   and push(@command, $self->dbname);
+    # No need to pass the database name PGDATABASE will be set
+    #  if a 'dbname' connection parameter was provided
 
-    $self->_run_command(command => [@command])
-        or croak 'error running command';
+    $self->_run_command(command => [@command],
+                        error   => 'error creating database');
 
     return 1;
 }
@@ -360,20 +649,15 @@ sub run_file {
     croak 'Must specify file' unless defined $args{file};
     croak 'Specified file does not exist' unless -e $args{file};
 
-    local $ENV{PGPASSWORD} = $self->password if defined $self->password;
-
     # Build command
-    my @command = ('psql', '--set=ON_ERROR_STOP=on', '-f', $args{file});
-    defined $self->username and push(@command, '-U', $self->username);
-    defined $self->host     and push(@command, '-h', $self->host);
-    defined $self->port     and push(@command, '-p', $self->port);
-    defined $self->dbname   and push(@command, $self->dbname);
+    my @command =
+        ($helper_paths{psql}, '--set=ON_ERROR_STOP=on', '-f', $args{file});
 
     my $result = $self->_run_command(
         command    => [@command],
         errlog     => $args{errlog},
         stdout_log => $args{stdout_log},
-    ) or croak 'error running command';
+        error      => "error running file '$args{file}'");
 
     return $result;
 }
@@ -426,19 +710,15 @@ sub backup {
     $self->{stderr} = undef;
     $self->{stdout} = undef;
 
-    local $ENV{PGPASSWORD} = $self->password if defined $self->password;
     my $output_filename = $self->_generate_output_filename(%args);
 
-    my @command = ('pg_dump', '-f', $output_filename);
-    defined $self->username and push(@command, '-U', $self->username);
-    defined $self->host     and push(@command, '-h', $self->host);
-    defined $self->port     and push(@command, '-p', $self->port);
+    my @command = ($helper_paths{pg_dump}, '-f', $output_filename);
     defined $args{compress} and push(@command, '-Z', $args{compress});
     defined $args{format}   and push(@command, "-F$args{format}");
-    defined $self->dbname   and push(@command, $self->dbname);
 
-    $self->_run_command(command => [@command])
-        or $self->_unlink_file_and_croak($output_filename, 'error running pg_dump command');
+    $self->_run_command(command => [@command],
+                        unlink  => [$output_filename],
+                        error   => 'error running pg_dump command');
 
     return $output_filename;
 }
@@ -484,13 +764,11 @@ sub backup_globals {
     local $ENV{PGPASSWORD} = $self->password if defined $self->password;
     my $output_filename = $self->_generate_output_filename(%args);
 
-    my @command = ('pg_dumpall', '-g', '-f', $output_filename);
-    defined $self->username and push(@command, '-U', $self->username);
-    defined $self->host     and push(@command, '-h', $self->host);
-    defined $self->port     and push(@command, '-p', $self->port);
+    my @command = ($helper_paths{pg_dumpall}, '-g', '-f', $output_filename);
 
-    $self->_run_command(command => [@command])
-        or $self->_unlink_file_and_croak($output_filename, 'error running pg_dumpall command');
+    $self->_run_command(command => [@command],
+                        unlink  => [$output_filename],
+                        error   => 'error running pg_dumpall command');
 
     return $output_filename;
 }
@@ -533,19 +811,15 @@ sub restore {
     return $self->run_file(%args)
            if not defined $args{format} or $args{format} eq 'p';
 
-    local $ENV{PGPASSWORD} = $self->password if defined $self->password;
-
     # Build command options
-    my @command = ('pg_restore', '--verbose', '--exit-on-error');
-    defined $self->dbname   and push(@command, '-d', $self->dbname);
-    defined $self->username and push(@command, '-U', $self->username);
-    defined $self->host     and push(@command, '-h', $self->host);
-    defined $self->port     and push(@command, '-p', $self->port);
+    my @command = ($helper_paths{pg_restore}, '--verbose', '--exit-on-error');
     defined $args{format}   and push(@command, "-F$args{format}");
+    defined $self->connect_data->{dbname}   and
+        push(@command, '-d', $self->connect_data->{dbname});
     push(@command, $args{file});
 
-    $self->_run_command(command => [@command])
-        or croak 'error running command';
+    $self->_run_command(command => [@command],
+                        error   => "error restoring from $args{file}");
 
     return 1;
 }
@@ -563,16 +837,11 @@ sub drop {
 
     croak 'No db name of this object' unless $self->dbname;
 
-    local $ENV{PGPASSWORD} = $self->password if $self->password;
+    my @command = ($helper_paths{dropdb});
+    push(@command, $self->connect_data->{dbname});
 
-    my @command = ('dropdb');
-    defined $self->username and push(@command, '-U', $self->username);
-    defined $self->host     and push(@command, '-h', $self->host);
-    defined $self->port     and push(@command, '-p', $self->port);
-    push(@command, $self->dbname);
-
-    $self->_run_command(command => [@command])
-        or croak 'error running command';
+    $self->_run_command(command => [@command],
+                        error   => 'error dropping database');
 
     return 1;
 }
@@ -640,7 +909,7 @@ L<http://search.cpan.org/dist/PGObject-Util-DBAdmin/>
 
 =head1 LICENSE AND COPYRIGHT
 
-Copyright 2014-2018 Chris Travers.
+Copyright 2014-2020 Chris Travers.
 
 This program is distributed under the (Revised) BSD License:
 L<http://www.opensource.org/licenses/BSD-3-Clause>
